/* r-io.c: input and output reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-io.c,v 2.12 2007/10/11 10:05:23 acau Exp $
 */


#include "system.h"




/***************************\
*                           *
*  input-output statements  *
*                           *
\***************************/

/* INPUT e */
void reduce_INPUT(e)
NODE1PTR e;
{
    char   name[MAXFILENAMESIZE];
    char c;
    FILE * fp = NULL,
         *openfile ();
    NODEPTR val;
    MEMORYPTR loc = reduce_ref (e -> c1);
    if (Ready) {
        output_state_number;
        fputs("% ", stdout); /* AC: added ? */
	output_expn (e -> c1, -1);
        fputs ("=?\n", stdout); /* AC: added ? */
        fflush(stdout);
        if (strcmp (InFile->str, "stdin") == 0) {
        val = reduce_val (parse (stdin));
        }
        else {
            if (Fpcur == NULL) {
                warn0 ("cannot open file will use stdin");
                val = reduce_val (parse (stdin));
            }
            else{
                val = parse (Fpcur);
                if (val == EndOfFile) {
                    warn0 ("end of file will use stdin");
                    val = reduce_val (parse (stdin));
                }
                else {
                    val = reduce_val (val);
                }
            }
        }
        
        
        
        if (Ready) {
	    if (!AssignAheadFlag && is_next (loc))
	        result (assign (e -> c1, val));
	    else {
		assign_val (loc, val);
	        result (True);
	    }
	}
    }
}

/* OUTPUT e1 */
void reduce_OUTPUT(e)
NODE1PTR e;
{
    NODEPTR e1 = reduce_ref_and_fetch (e -> c1);
    
    if ((Ready && val_defined (e1)) || ForceOutput) {
        output_state_number;
        if (strcmp (OutFile->str, "stdout") != 0) {
            newline;
        }
        foutput_expn (e -> c1, -1);
        foutput_str ("=");
        if (Ready)
            foutput_val (e1);
        else
            foutput_char ('?');
        fnewline;
        result (True);
    }
    Ready = (Ready && val_defined (e1)) || ForceOutput;
}

void reduce_PEEK(e)
NODE1PTR e;
{
    NODEPTR e1;
    foutput_expn (e -> c1, -1);
    foutput_str ("=");
    e1 = reduce_ref_and_fetch (e -> c1);
    if (Ready)
        foutput_val (e1);
    else
        foutput_char ('?');
    fnewline;
    Ready = T;
    result (True);
}

/* Output a value according to a given print format. */
/* FORMAT elist */
void reduce_FORMAT(e)
NODE1PTR e;
{
    NODE2PTR elist = e -> c1;
    NODEPTR fmt;
    NODEPTR item;
    MEMORYPTR base_loc,
            loc;
    int     field_size;
    char   *fmt_ptr;
    char   name[MAXFILENAMESIZE];
    FILE * fp = NULL,
         *openfile ();

    assert0 (elist != NULL, "format statement must contain format string");
    loc = base_loc = mem_block (STATE, elist_size (elist)) -> blk_base;
    Ready = T;
    while ((Ready || ForceOutput) && (elist != NULL)) {
        item = reduce_ref_and_fetch (elist -> c1);
        if (Ready) {
	    put_mem (curr_loc (loc), item);
	    Ready = val_defined (item);
	}
        elist = elist -> c2;
        loc += offset (1);
    }
    if (Ready || ForceOutput) {
        type_check (fmt = get_mem (loc = curr_loc (base_loc)), STRING);
        output_state_number;
        for (fmt_ptr = stringp (fmt) -> str; *fmt_ptr != '\0'; fmt_ptr += 1) {
            if (*fmt_ptr != '%') {
                if (strcmp (OutFile->str, "stdout") == 0) {
                    putchar (*fmt_ptr);
                }
                else {
                    if ((fp = openfile (name, Path->str, OutFile->str, FXOUT, "a")) == NULL) {
                        warn1 ("cannot open file \"%s\"",  OutFile->str);
                        warn0 ("will display to stdout");
                        putchar (*fmt_ptr);
                    }
                    else {
                        putc(*fmt_ptr,fp);
                        fclose (fp);
                    }
                }
            }
            else {
                CharsOutput = 0;
                field_size = 0;
                while (isdigit ((int)*++fmt_ptr)) {
                    field_size = field_size * 10 + *fmt_ptr - '0';
                }
                switch (*fmt_ptr) {
                    case 'd':
                    case 'D':
		        if (mem_defined (loc += offset (1))) {
			    type_check (item = get_mem (loc), INT);
			    foutput_int (intp (item) -> ival, field_size);
			} else 
			    foutputf_str ("?", field_size);
                        break;
                    case 's':
                    case 'S':
		        if (mem_defined (loc += offset (1))) {
                            type_check (item = get_mem (loc), STRING);
                            foutputf_str (stringp (item) -> str, field_size);
			} else 
			    foutputf_str ("?", field_size);
                        break;
                    case 't':
                    case 'T':
		        if (mem_defined (loc += offset (1)))
                            foutput_expn (item = get_mem (loc), -1);
			else 
			    foutputf_str ("?", field_size);
                        break;
                    case '#':
                        foutput_int ((int)StateNumber, field_size);
                        break;
                    default:
                        if (strcmp (OutFile->str, "stdout") == 0) {
                            putchar (*fmt_ptr);
                        }
                        else {
                            if ((fp = openfile (name, Path->str, OutFile->str, FXOUT, "a")) == NULL) {
                                warn1 ("cannot open file \"%s\"",  OutFile->str);
                                warn0 ("will display to stdout");
                                putchar (*fmt_ptr);
                            }
                            else {
                                putc(*fmt_ptr,fp);
                                fclose (fp);
                            }
                        }
                }
                fspaces (field_size - CharsOutput);
            }
        }
        result (True);
    }
}
